/*
  A simple Mandelbrot polynomial roots calculator.
  Adapted from an evadraw script.
  It saves the result in a file named "save.txt"
  (finally using templates wasn't really a good idea)
  compile with g++ and use -Ofast switch:
  g++ -Ofast -o Mandelroots.exe Mandelroots.cpp
  Usage:
  mandelroots period
  period should be between 1 and 26
  This program uses double floats. The float type can be changed in the declaration of mrc in main()
 */
#include <iostream>
#include <fstream>
#include <algorithm>
#include <vector>
#include <ctime>
#include <complex>
#include <omp.h>
#include <sys/time.h>

using namespace std;

//---------------------------------------------------------------------------------------------------
#define COMPLEX std::complex<T>

// Say F_{n}(c,z) is defined as:
// F_{n+1} = F_{n}² + c
// F_{0}(c,z) = z
// This function returns:
// - the number of iterations used: < MaxIter if bailing out and ==MaxIter otherwise
// - z = F_{MaxIter}(c,0) ; In case no bailout occure
// - dz = d/dc F_{MaxIter}(c,0)
template<typename T>
inline int MPoly_ZdZ(const COMPLEX c, const int MaxIter, COMPLEX& z, COMPLEX& dz)
{
    z=COMPLEX(0.,0.);// Of course this line can be removed. Then, the returned z_out = F_{MaxIter}(c,z_in)
    dz=COMPLEX(0.,0.);
    int i=0;
    for(; i<MaxIter && norm(z)<T(4.); i++){
        dz = T(2.) * z * dz + COMPLEX(1.);
        z  = z * z + c;
    }
    return i;
}

// Returns true if F_{MaxIter}([c],0) contains 0. that means that we MAY have a zero.
// if F_{MaxIter}([c],0) does not contain 0 then we are SURE there is no roots
// here [c] is the complex interval which is a disc of radius r centered at c.
// [c] = c + [e]r
// [e] = [0,1] exp(i [-inf, +inf]); it is the disc of radius 1 centered at 0.
// [e] can also be seen as a function: e(t,a)= t exp(i a) where t is in [0,1] and a is in [-inf, +inf]
// Here we use a kind of Reduced Affine Arithmetic (RAA) augmented by Taylor expansion... We set:
//   [zn] = F_{n}([c],0) = zn + [e]r dzn + 1/2 [e]² r² En
//   Where:
//     zn = F_{n}(c,0)
//     dzn= d/dc F_{n}(c,0)
//     En is a real number that contains the upper order error
//  Now we have
//    [zn1] = [zn]² + [c]
//          = (zn + [e]r dzn + 1/2 [e]² r² En)² + c + [e]r
//          = zn² + 2 [e]r zn dzn + 2/2 [e]² r² zn En + [e]²r² dzn² + 2/2 [e]^3 r^3 dzn En + 1/4 [e]^4 r^4 En² + c + [e]r
//          = (zn² + c) + (2 zn dzn + 1) [e]r + 1/2 [e]² r² (2 zn En + 2 dzn² + 2 [e]r dzn En + 1/2 [e]² r² En²)
//          = zn1 + dzn1 [e]r + 1/2 [e]² r² [En1]
//  Reducing [En1]:
//    En1 = 2 |zn| En + 2 |dzn|² + 2 [e]r |dzn| En + 1/2 [e]² r² En²
//    En1 = 2 (|zn| En + |dzn|² + r |dzn| En + 1/4 r² En²)
// One can also use higher and lower taylor expansion orders. For ex:
//    zn + [e]r En ; Lower order
//    zn + [e]r dzn + 1/2 [e]² r² ddzn + 1/6 [e]^3 r^3 En
//    ... etc.
// And yes, it really looks like "perturbation theory" of superfractalthing! for which it may be useful :)
//
// The drawback of this methos is that we have to compute absolute values of complex numbers which requires sqrt()s
// but it doesn't suffer from (catastrophic) interval expansion of regular IA when dowing multiplications.
// TODO: Implement an RAA-Taylor class... Anyway, for this application we won't use it too much. :/
template<typename T>
inline bool MFDiscContainsZero(const COMPLEX c, const T r, const int MaxIter)
{
    COMPLEX z = COMPLEX(0.);
    COMPLEX dz= COMPLEX(0.);
    T rz = T(0.), rdz= T(0.), Ei= T(0.), rr= T(0.);
    for(int i=0; i<MaxIter &&
                 rz-rr < T(2.) && /*bailout if [z] completly outside Disc((0,0),2)*/
                 rz-rr > T(-2.) /*or if [z] contains Disc((0,0),2)*/
               ; i++){
        Ei = T(2.) * (rdz * rdz + (rz + (rdz + T(0.25) * r * Ei) * r) * Ei);
        dz = T(2.) * z * dz + COMPLEX(1.);
        z  = z * z + c;
        rdz = abs(dz);
        rz = abs(z);
        rr = r * (rdz + T(0.5) * r * Ei); // Radius of uncertainty disc around z_i
    }
    if(rz - rr <= T(0.)) return true;//if [zn] contains 0 return true
    return false;
}

//same as above but with :
// [zn] = F_{n}([c],0) = zn + [e]r dzn + [e]² r² En
template<typename T>
bool MFDiscContainsZero1(const COMPLEX c, const T r, const int MaxIter)
{
    COMPLEX z = COMPLEX(0.);
    COMPLEX dz= COMPLEX(0.);
    T rz = T(0.), rdz= T(0.), Ei= T(0.), rr= T(0.);
    for(int i=0; i<MaxIter && rz-rr < T(2.) && rz-rr > T(-2.); i++){//bail out when [zn] is entirely outside 2... or if it covers everything
        Ei = rdz * rdz + (T(2.) * rz + r * (T(2.) * rdz + r * Ei)) * Ei;
        dz = T(2.) * z * dz + COMPLEX(1.);
        z  = z * z + c;
        rdz = abs(dz);
        rz = abs(z);
        rr = r * (rdz + r * Ei); // Radius of uncertainty disc around z_i
    }
    if(rz - rr <= T(0.)) return true;//if [zn] contains 0 return true
    return false;
}
//-------------------------------------------------------------------------------------------------------
//these are used for sorting the array containing the roots
template<typename T>
struct CompRup{
    bool operator()(const complex<T>& c0, const complex<T>& c1){
        return c0.real() < c1.real();
    }
};

template<typename T>
struct CompIup{
    bool operator()(const complex<T>& c0, const complex<T>& c1){
        return c0.imag() < c1.imag();
    }
};

//The class that computes the roots
template<typename T>
class MRootsCalc
{
public:
    MRootsCalc(int MaxIter){
        m_MaxIter = MaxIter;// it's actually the order of the Mandelbrot polynomial
        // choosing the FP accuracy and setting the maximal MaxIter wrt the floating point type used
        // MaxIter should less than bits/2-1. This is due to rounding errors... it will work but will miss some roots when maxIter > bits/2-1
        switch (sizeof(T)){
            case 4 : {m_epsilon = pow(2., -24); m_MaxIter = min(11,m_MaxIter); break;}
            case 8 : {m_epsilon = pow(2., -54); m_MaxIter = min(26,m_MaxIter); break;}
            case 10:
            case 12: {m_epsilon = pow(2., -64); m_MaxIter = min(31,m_MaxIter); break;} //  it's slow. not using FPU?
            case 16: {m_epsilon = pow(2., -64); m_MaxIter = min(33,m_MaxIter); break;} //__float128. doesn't work ... need more info
        default : m_epsilon = 1.e-15;// should not happen :/
        }
        m_eps = pow(2., -2*(m_MaxIter - 1));
        m_nbrRoots = 1 << (m_MaxIter - 1);
        m_tab.reserve(2*m_nbrRoots);//because subdivision process may report a root more than once
    }

    inline void findRoots(){
        cout << "tab size: " << m_tab.capacity() << endl;
        cout << "nbrRoots (expected) = " << m_nbrRoots << endl;
        cout << "Searchig for roots..." << endl;
        //subdivide
        m_cIndex=0;
        findRootsRecurse(complex<T>(0.,0.), T(2.));//this is the stupid way
        /* faster way :
            findRootsRecurse(complex<T>(-1.,1.), T(1.));
            findRootsRecurse(complex<T>(1.,1.), T(1.));
           then if c is a non real root, cs=complex(c.real(),-c.imag()) is also a root.
        */
        cout << "roots found (with duplicates): " << m_tab.size() << endl;
        //filter
        cout << "Removing duplicates..." << endl;
        filter();
        cout << "roots found: " << m_tab.size() << endl;
        cout << m_tab.size() << endl;
    }

    void printThem(){
        switch (sizeof(T)){
            case 4 : {cout.precision(8); break;}
            case 8 : {cout.precision(16); break;}
            case 10:
            case 12: {cout.precision(20); break;}
            case 16: {cout.precision(25); break;} //__float128. doesn't work ... need more info
        default : cout.precision(16);// should not happen :/
        }
        cout.setf(ios::fixed);
        for (int i = 0; i < m_tab.size() /*m_cIndex*/; i++){
            cout << m_tab[i].real() << "\t" << m_tab[i].imag() << endl;
        }
    }

    void saveThem(){
        ofstream SaveFile("save.txt");

        switch (sizeof(T)){
            case 4 : {SaveFile.precision(8); break;}
            case 8 : {SaveFile.precision(16); break;}
            case 10:
            case 12: {SaveFile.precision(20); break;}
            case 16: {SaveFile.precision(25); break;} //__float128. doesn't work ... need more info
        default : SaveFile.precision(16);// should not happen :/
        }
        SaveFile.setf(ios::fixed);
        for (unsigned int i = 0; i < m_tab.size() /*m_cIndex*/; i++){
            SaveFile << m_tab[i].real() << "\t" << m_tab[i].imag() << endl;
        }
    }

private:
    inline void filter(){
        //sort in real direction
        std::sort(m_tab.begin(),m_tab.end(),CompRup<T>());
        //remove duplicates
        clean();
        //sort in imaginary direction
        std::sort(m_tab.begin(),m_tab.end(),CompIup<T>());
        //remove duplicates
        clean();
        //sort in real direction
        std::sort(m_tab.begin(),m_tab.end(),CompRup<T>());
    }

    inline void clean(){
        int i0 = 0;
        int i1 = 1;
        while (i1 < m_cIndex){
            if(std::max(abs(m_tab[i1].real()-m_tab[i0].real()), abs(m_tab[i1].imag()-m_tab[i0].imag())) < m_eps){
                i1++;
            } else {
                i0++;
                m_tab[i0] = m_tab[i1];
                i1++;
            }
        }
        m_cIndex = i0+1;
        m_tab.resize(m_cIndex);
    }

    inline complex<T> refine(complex<T> c){
        //use Newton method
        complex<T> z, dz;
        for(int i=0; i<20; i++){
            int mi=MPoly_ZdZ(c, m_MaxIter, z, dz);
            if(mi<m_MaxIter) cout << "boum";
            complex<T> delta = z / dz;
            c -= delta;
            //if(norm(delta) < m_epsilon*m_epsilon) break;
            if((max(abs(delta.real()),abs(delta.imag()))) < T(0.5)*m_epsilon) break;
            //if(abs(delta/c) < m_epsilon) break;
        }
        return c;
    }

    inline void store(complex<T> cc){
        if (m_cIndex>0){//first time filtering
            complex<T> cc1 = m_tab[m_cIndex -1];
            if(std::max(abs(cc1.real()-cc.real()), abs(cc1.imag()-cc.imag())) < m_eps) return;
        }
        m_tab.push_back(cc);
        m_cIndex++;
    }

		void findRootsRecurse(complex<T> cc /*center of square*/, T hw /*half width of square*/, int depth = 0){
        //verify if there is no roots in the disc (cc, hw*sqrt(2.))
        if( ! MFDiscContainsZero(cc, hw * T(1.42), m_MaxIter) ) return;
        //verify if there is only one root. if true, report it in m_tab.
        //One way of doing this is to compute an interval of the derivative of F_{n}(c,0) wrt c.
        //if this interval doesn't contain 0 then there is only one root.
        //but we use a simpler test. we know that the minimal separation between roots is m_eps.
        if(hw<m_eps){
            //we know there is only one root. refine then save it.
            cc = refine(cc);
            store(cc);
            return;
        }
        //Many possible roots -> subdivide
				hw *= T(0.5);
				if(depth <= 2)
				{
					#pragma omp parallel for
					for(int i = 0; i < 4; i++){
						findRootsRecurse(cc + complex<T>(-hw + (i / 2) * hw,-hw + (i % 2) * hw), hw, depth + 1);
					}
				}
				else{
					findRootsRecurse(cc + complex<T>(-hw,-hw),hw, depth + 1);
					findRootsRecurse(cc + complex<T>( hw,-hw),hw, depth + 1);
					findRootsRecurse(cc + complex<T>( hw, hw),hw, depth + 1);
					findRootsRecurse(cc + complex<T>(-hw, hw),hw, depth + 1);
				}
    }
    int m_MaxIter;
    int m_nbrRoots;
    int m_cIndex;
    T m_eps, m_epsilon;
    vector< complex<T> > m_tab;
};
//-------------------------------------------------------------------------------------------------------

int main(int argc, char* argv[])
{
		omp_set_nested(1); // allows nested openmp splitting in recursive function calls
    int period=10;
    if(argc>1){
        unsigned int n = atoi(argv[1]);//what if argv[1] is not a number
        period = n;
    }

    MRootsCalc<double> mrc(period);
		struct timeval start, end;
		gettimeofday(&start, NULL);

		// benchmark code
    mrc.findRoots();
		gettimeofday(&end, NULL);
    //mrc.printThem();
		double delta = ((end.tv_sec  - start.tv_sec) * 1000000u +
						 end.tv_usec - start.tv_usec) / 1.e6;
		cout << "time taken : " << double(delta) << " s"<< endl;
    cout << "Saving results to save.txt" << endl;
    mrc.saveThem();

    return 0;
}

